/******************************************************************************
 *    Copyright (C) 2006 by JOSEPH Nicolas                                    *
 *    gege2061@redaction-developpez.com                                       *
 *                                                                            *
 *    This program is free software; you can redistribute it and/or modify    *
 *    it under the terms of the GNU General Public License as published by    *
 *    the Free Software Foundation; either version 2 of the License, or       *
 *    (at your option) any later version.                                     *
 *                                                                            *
 *    This program is distributed in the hope that it will be useful,         *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *    GNU General Public License for more details.                            *
 *                                                                            *
 *    You should have received a copy of the GNU General Public License       *
 *    along with this program; if not, write to the                           *
 *    Free Software Foundation, Inc.,                                         *
 *    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.               *
 ******************************************************************************/

#include "gmarkup-dom.h"

#include <ctype.h>

#include <glib/gfileutils.h>
#include <glib/gmarkup.h>
#include <glib/gmem.h>
#include <glib/gmessages.h>
#include <glib/gstrfuncs.h>

/* macros =================================================================== */
/* constants ================================================================ */
/* types ==================================================================== */
typedef struct _GMarkupDomContext GMarkupDomContext;

/* structures =============================================================== */
struct _GMarkupDomContext
{
  gint level;
  GMarkupDomNode *root;
  GMarkupDomNode *last_root;
  GMarkupDomNode *last_node;
  GMarkupDomNode *current;
};

/* private variables ======================================================== */
/* private functions ======================================================== */
static int str_isspace (const gchar *s)
{
  if (!s) return 0;
  for (; *s && isspace (*s); s++);
  return !*s;
}

void xml_start_element (GMarkupParseContext *context, const gchar *element_name,
                        const gchar **attribute_names,
                        const gchar **attribute_values, gpointer user_data,
                        GError **error)
{
  gint i;
  GMarkupDomContext *dom_context = user_data;
  GMarkupDomNode *node = NULL;

  g_return_if_fail (dom_context != NULL);

  node = g_malloc (sizeof (*node));
  node->name = NULL;
  node->content = NULL;
  node->level = 0;
  node->attributs = NULL;
  node->parent = NULL;
  node->child = NULL;
  node->next = NULL;
  node->prev = NULL;

  dom_context->level++;
  dom_context->current = node;
  if (dom_context->root == NULL)
  {
    dom_context->root = node;
  }

  node->name = g_strdup (element_name);
  node->level = dom_context->level;

  /* previous - next */
  if (dom_context->last_node != NULL
      && dom_context->last_node->level == node->level)
  {
    node->prev = dom_context->last_node;
    dom_context->last_node->next = node;
  }

  /* Copy attributs */
  for (i = 0; attribute_names[i] != NULL; i++)
  {
  }
  node->attributs = g_malloc (sizeof (*node->attributs) * (i + 1));
  for (i = 0; attribute_names[i] != NULL; i++)
  {
    node->attributs[i].name = g_strdup (attribute_names[i]);
    node->attributs[i].value = g_strdup (attribute_values[i]);
  }
  node->attributs[i].name = NULL;
  node->attributs[i].value = NULL;

  /* parent - child */
  node->parent = dom_context->last_root;
  if (dom_context->last_root && dom_context->last_root->child == NULL)
  {
    dom_context->last_root->child = node;
  }
  dom_context->last_root = node;

  /* Unused parameters */
  (void)context;
  (void)error;
}

static void xml_end_element (GMarkupParseContext *context,
                             const gchar *element_name, gpointer user_data,
                             GError **error)
{
  GMarkupDomContext *dom_context = user_data;

  g_return_if_fail (dom_context != NULL);
  g_return_if_fail (dom_context->current != NULL);

  dom_context->level--;
  dom_context->last_root = dom_context->current->parent;
  dom_context->last_node = dom_context->current;
  dom_context->current = dom_context->current->parent;

  /* Unused parameters */
  (void)context;
  (void)element_name;
  (void)error;
}

static void xml_text (GMarkupParseContext *context, const gchar *text,
                      gsize text_len, gpointer user_data, GError **error)
{
  GMarkupDomContext *dom_context = user_data;

  g_return_if_fail (dom_context != NULL);
  g_return_if_fail (dom_context->current != NULL);

  if (dom_context->current->content == NULL && !str_isspace (text))
  {
    dom_context->current->content = g_strdup (text);
  }

  /* Unused parameters */
  (void)context;
  (void)text_len;
  (void)error;
}

/* internal public functions ================================================ */
/* entry points ============================================================= */
/**
 * g_markup_dom_new:
 * @file_name: name of a file to parse contents from.
 * @error: return location for a #GError, or %NULL.
 *
 * Create a dom tree of @filename content.
 *
 * Return value: a new #GMarkupDomNode.
 **/
GMarkupDomNode *g_markup_dom_new (const gchar *filename, GError **error)
{
  GMarkupParseContext *markup_parse_context = NULL;
  GMarkupDomContext context = {0, NULL, NULL, NULL, NULL};

  g_return_val_if_fail (filename != NULL, context.root);

  {
    GMarkupParser markup_parser;

    markup_parser.start_element = xml_start_element;
    markup_parser.end_element = xml_end_element;
    markup_parser.text = xml_text;
    markup_parser.passthrough = NULL;
    markup_parser.error = NULL;
    markup_parse_context = g_markup_parse_context_new (&markup_parser, 0,
                                                       &context, NULL);
  }
  {
    gchar *text = NULL;
    gsize length = -1;

    g_file_get_contents (filename, &text, &length, error);
    if (text != NULL)
    {
      g_markup_parse_context_parse (markup_parse_context, text, length, error);
      g_free (text), text = NULL;
    }
    g_free (markup_parse_context), markup_parse_context = NULL;
  }
  return context.root;
}

/**
 * g_markup_dom_free:
 * @node: a #GMarkupDomNode.
 *
 * Frees a #GMarkupDomNode.
 **/
void g_markup_dom_free (GMarkupDomNode *node)
{
  if (node != NULL)
  {
    gint i;

    g_free (node->name), node->name = NULL;
    g_free (node->content), node->content = NULL;
    for (i = 0; node->attributs[i].name; i++)
    {
      g_free (node->attributs[i].name), node->attributs[i].name = NULL;
      g_free (node->attributs[i].value), node->attributs[i].value = NULL;
    }
    g_free (node->attributs), node->attributs = NULL;
    node->parent = NULL;
    node->prev = NULL;
    g_markup_dom_free (node->child), node->child = NULL;
    g_markup_dom_free (node->next), node->next = NULL;
    g_free (node), node = NULL;
  }
}

/* public variables ========================================================= */
